# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_batch_json


@pytest.fixture(scope="function")
def valid_batch_definition():
    return {
        "type": "batch",
        "tasks": [{"order_id": 1, "task_id": "sleep-300"}],
    }


def test_batch(valid_batch_definition):
    json_data = valid_batch_definition

    assert validate_batch_json(json_data)


def test_batch_missing_tasks(valid_batch_definition):
    json_data = valid_batch_definition

    del json_data["tasks"]
    assert not validate_batch_json(json_data)


def test_batch_timeout(valid_batch_definition):
    json_data = valid_batch_definition

    json_data["tasks"][0]["timeout"] = 10
    assert validate_batch_json(json_data)

    json_data["tasks"][0]["timeout"] = 0
    assert not validate_batch_json(json_data)

    json_data["tasks"][0]["timeout"] = -1
    assert not validate_batch_json(json_data)


def test_batch_continue_on_fail(valid_batch_definition):
    json_data = valid_batch_definition

    json_data["tasks"][0]["continue_on_fail"] = True
    assert validate_batch_json(json_data)

    json_data["tasks"][0]["continue_on_fail"] = False
    assert validate_batch_json(json_data)


def test_batch_retry_on_rerun(valid_batch_definition):
    json_data = valid_batch_definition

    json_data["tasks"][0]["retry_on_rerun"] = True
    assert validate_batch_json(json_data)

    json_data["tasks"][0]["retry_on_rerun"] = False
    assert validate_batch_json(json_data)



# pylint: skip-file
# ruff: noqa
import json
import os

from pytest_shell import fs

from opentaskpy.config.loader import ConfigLoader
from opentaskpy.taskhandlers import transfer
from tests.fixtures.ssh_clients import *  # noqa: F403

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

# Create a task definition
email_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*emailhandler.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "recipients": ["test@example.com", "test1@example.com"],
            "subject": "Test Email Subject",
            "protocol": {
                "name": "email",
                "credentials": {
                    "username": "{{ lookup('file', path='/tmp/smtp_username') }}",
                    "password": "{{ lookup('file', path='/tmp/smtp_password') }}",
                },
                "sender": "Test Sender <test@example.com>",
            },
        },
    ],
}

email_dummy_source_task_definition = {
    "type": "transfer",
    "source": {
        "protocol": {"name": "dummy"},
    },
    "destination": [
        {
            "recipients": ["test@example.com"],
            "subject": "Test Email Subject DUMMY SOURCE",
            "message": "This has not an attachment, but a custom body",
            "protocol": {
                "name": "email",
                "credentials": {
                    "username": "{{ lookup('file', path='/tmp/smtp_username') }}",
                    "password": "{{ lookup('file', path='/tmp/smtp_password') }}",
                },
                "sender": "Test Sender <test@example.com>",
            },
        },
    ],
}

email_html_dummy_source_task_definition = {
    "type": "transfer",
    "source": {
        "protocol": {"name": "dummy"},
    },
    "destination": [
        {
            "recipients": ["test@example.com"],
            "subject": "Test Email Subject DUMMY SOURCE",
            "messageContentType": "text/html",
            "message": (
                'This has not attachment, but a <b>custom body</b> with some <p style="color: red;">red text in a paragraph.</p>'
            ),
            "protocol": {
                "name": "email",
                "credentials": {
                    "username": "{{ lookup('file', path='/tmp/smtp_username') }}",
                    "password": "{{ lookup('file', path='/tmp/smtp_password') }}",
                },
                "sender": "Test Sender <test@example.com>",
            },
        },
    ],
}

email_plain_dummy_source_task_definition = {
    "type": "transfer",
    "source": {
        "protocol": {"name": "dummy"},
    },
    "destination": [
        {
            "recipients": ["test@example.com"],
            "subject": "Test Email Subject DUMMY SOURCE",
            "messageContentType": "text/plain",
            "message": "This has not attachment, but a plain message",
            "protocol": {
                "name": "email",
                "credentials": {
                    "username": "{{ lookup('file', path='/tmp/smtp_username') }}",
                    "password": "{{ lookup('file', path='/tmp/smtp_password') }}",
                },
                "sender": "Test Sender <test@example.com>",
            },
        },
    ],
}

email_task_definition_message_content_filename = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*emailhandler.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "recipients": ["test@example.com"],
            "subject": "Test Email Subject",
            "protocol": {
                "name": "email",
                "credentials": {
                    "username": "{{ lookup('file', path='/tmp/smtp_username') }}",
                    "password": "{{ lookup('file', path='/tmp/smtp_password') }}",
                },
                "sender": "Test Sender <test@example.com>",
            },
            "messageContentFilename": "/tmp/email-message.txt",
        },
    ],
}


def test_remote_handler():
    # Validate that given a transfer with email protocol, that we get a remote handler of type EmailTransfer

    transfer_obj = transfer.Transfer(None, "email-basic", email_task_definition)

    transfer_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert transfer_obj.source_remote_handler.__class__.__name__ == "SSHTransfer"

    # dest_remote_handler should be an array
    assert isinstance(transfer_obj.dest_remote_handlers, list)
    assert len(transfer_obj.dest_remote_handlers) == 1
    #  of SSHTransfer objects
    assert transfer_obj.dest_remote_handlers[0].__class__.__name__ == "EmailTransfer"


def test_remote_handler_vars(env_vars):
    # Load the global config
    config_loader = ConfigLoader("test/cfg")
    global_variables = config_loader.get_global_variables()
    global_variables["global_protocol_vars"] = [
        {"name": "email", "smtp_port": 587, "smtp_server": "smtp.gmail.com"}
    ]

    transfer_obj = transfer.Transfer(
        global_variables, "email-basic", email_task_definition
    )
    transfer_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert transfer_obj.source_remote_handler.__class__.__name__ == "SSHTransfer"

    # dest_remote_handler should be an array
    assert isinstance(transfer_obj.dest_remote_handlers, list)
    assert len(transfer_obj.dest_remote_handlers) == 1
    #  of SSHTransfer objects
    assert transfer_obj.dest_remote_handlers[0].__class__.__name__ == "EmailTransfer"

    # Check that the transfer object has an SMTP server config from the global config
    assert (
        transfer_obj.dest_remote_handlers[0].protocol_vars["smtp_server"]
        == "smtp.gmail.com"
    )


def test_email_transfer(env_vars, setup_ssh_keys, root_dir):

    import logging

    import opentaskpy.otflogging

    logger = opentaskpy.otflogging.init_logging(
        __name__, "email-transfer", level=logging.DEBUG, override_root_logger=True
    )

    # In GitHub Actions, the variables we need are in the environment
    # Pull those and write them to the config files first
    if os.getenv("GITHUB_ACTIONS"):
        # Get SMTP_USERNAME and SMTP_PASSWORD from environment and write them to files under /tmp
        fs.create_files(
            [
                {
                    "/tmp/smtp_username": {
                        "content": os.getenv("SMTP_USERNAME"),
                    }
                },
                {
                    "/tmp/smtp_password": {
                        "content": os.getenv("SMTP_PASSWORD"),
                    }
                },
            ]
        )

    # Create a file to transfer
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/emailhandler.txt": {"content": "test1234"}}]
    )

    # Write the email_task_definition to a file which we will read in to resolve the templated values for username and password
    task_definition_file = f"{root_dir}/cfg/transfers/email-transfer.json"
    # Delete the file if it exists
    if os.path.exists(task_definition_file):
        os.remove(task_definition_file)

    fs.create_files(
        [{task_definition_file: {"content": json.dumps(email_task_definition)}}]
    )

    # Load the global config
    config_loader = ConfigLoader("test/cfg")
    global_variables = config_loader.get_global_variables()
    global_variables["global_protocol_vars"] = [
        {"name": "email", "smtp_port": 587, "smtp_server": "smtp.gmail.com"}
    ]

    # Load the task definition using the config_loader
    imported_task_def = config_loader.load_task_definition("email-transfer")
    os.remove(task_definition_file)

    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    assert transfer_obj.run()

    # Check that files have been tidied up on the worker
    assert not os.path.exists(transfer_obj.local_staging_dir)

    # Now test the dummy source
    # Write the email_task_definition to a file which we will read in to resolve the templated values for username and password
    task_definition_file = f"{root_dir}/cfg/transfers/email-transfer-dummy.json"
    # Delete the file if it exists
    if os.path.exists(task_definition_file):
        os.remove(task_definition_file)

    fs.create_files(
        [
            {
                task_definition_file: {
                    "content": json.dumps(email_dummy_source_task_definition)
                }
            }
        ]
    )

    imported_task_def = config_loader.load_task_definition("email-transfer-dummy")
    os.remove(task_definition_file)

    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer-dummy", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    assert transfer_obj.run()


def test_email_transfer_message_content_filename(env_vars, setup_ssh_keys, root_dir):

    import logging

    import opentaskpy.otflogging

    logger = opentaskpy.otflogging.init_logging(
        __name__, "email-transfer-message-content-filename", level=logging.DEBUG
    )

    # In GitHub Actions, the variables we need are in the environment
    # Pull those and write them to the config files first
    if os.getenv("GITHUB_ACTIONS"):
        # Get SMTP_USERNAME and SMTP_PASSWORD from environment and write them to files under /tmp
        fs.create_files(
            [
                {
                    "/tmp/smtp_username": {
                        "content": os.getenv("SMTP_USERNAME"),
                    }
                },
                {
                    "/tmp/smtp_password": {
                        "content": os.getenv("SMTP_PASSWORD"),
                    }
                },
            ]
        )

    # Create a file to transfer
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/emailhandler.txt": {"content": "test1234"}}]
    )

    # Write the email_task_definition to a file which we will read in to resolve the templated values for username and password
    task_definition_file = (
        f"{root_dir}/cfg/transfers/email-transfer-message-content-filename.json"
    )
    # Delete the file if it exists
    if os.path.exists(task_definition_file):
        os.remove(task_definition_file)

    fs.create_files(
        [
            {
                task_definition_file: {
                    "content": json.dumps(
                        email_task_definition_message_content_filename
                    )
                },
            },
            {
                "/tmp/email-message.txt": {
                    "content": "This is the email message from a file"
                },
            },
        ]
    )

    # Load the global config
    config_loader = ConfigLoader("test/cfg")
    global_variables = config_loader.get_global_variables()
    global_variables["global_protocol_vars"] = [
        {"name": "email", "smtp_port": 587, "smtp_server": "smtp.gmail.com"}
    ]

    # Load the task definition using the config_loader
    imported_task_def = config_loader.load_task_definition(
        "email-transfer-message-content-filename"
    )
    os.remove(task_definition_file)

    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer-message-content-filename", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    assert transfer_obj.run()

    # Check that files have been tidied up on the worker
    assert not os.path.exists(transfer_obj.local_staging_dir)

    # Check that the message file has been deleted too
    assert not os.path.exists("/tmp/email-message.txt")

    # Write the email message file again, and this time send the email but explicitl specify not to delete the file
    fs.create_files(
        [
            {
                "/tmp/email-message.txt": {
                    "content": "This is the email message from a file"
                },
            },
        ]
    )

    imported_task_def["destination"][0]["deleteContentFileAfterTransfer"] = False
    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer-message-content-filename", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    assert transfer_obj.run()

    # Check that files have been tidied up on the worker
    assert not os.path.exists(transfer_obj.local_staging_dir)

    # Check that the message file has not been deleted
    assert os.path.exists("/tmp/email-message.txt")

    # Add a message to the task definition and validate that it failed because the message and messageContentFilename are both defined
    imported_task_def["destination"][0]["message"] = "This is the email message"
    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer-message-content-filename", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    with pytest.raises(Exception) as ex:
        transfer_obj.run()


def test_email_content_type(env_vars, setup_ssh_keys, root_dir):

    import logging

    import opentaskpy.otflogging

    logger = opentaskpy.otflogging.init_logging(
        __name__,
        "email-transfer-content-type",
        level=logging.DEBUG,
        override_root_logger=True,
    )

    # In GitHub Actions, the variables we need are in the environment
    # Pull those and write them to the config files first
    if os.getenv("GITHUB_ACTIONS"):
        # Get SMTP_USERNAME and SMTP_PASSWORD from environment and write them to files under /tmp
        fs.create_files(
            [
                {
                    "/tmp/smtp_username": {
                        "content": os.getenv("SMTP_USERNAME"),
                    }
                },
                {
                    "/tmp/smtp_password": {
                        "content": os.getenv("SMTP_PASSWORD"),
                    }
                },
            ]
        )

    # Now test the dummy source
    # Write the email_task_definition to a file which we will read in to resolve the templated values for username and password
    task_definition_file = f"{root_dir}/cfg/transfers/email-transfer-dummy-html.json"
    # Delete the file if it exists
    if os.path.exists(task_definition_file):
        os.remove(task_definition_file)

    fs.create_files(
        [
            {
                task_definition_file: {
                    "content": json.dumps(email_html_dummy_source_task_definition)
                }
            }
        ]
    )

    config_loader = ConfigLoader("test/cfg")
    global_variables = config_loader.get_global_variables()
    global_variables["global_protocol_vars"] = [
        {"name": "email", "smtp_port": 587, "smtp_server": "smtp.gmail.com"}
    ]

    imported_task_def = config_loader.load_task_definition("email-transfer-dummy-html")
    os.remove(task_definition_file)

    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer-dummy-html", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    assert transfer_obj.run()

    # Now test the dummy source
    # Write the email_task_definition to a file which we will read in to resolve the templated values for username and password
    task_definition_file = f"{root_dir}/cfg/transfers/email-transfer-dummy-plain.json"
    # Delete the file if it exists
    if os.path.exists(task_definition_file):
        os.remove(task_definition_file)

    fs.create_files(
        [
            {
                task_definition_file: {
                    "content": json.dumps(email_plain_dummy_source_task_definition)
                }
            }
        ]
    )

    config_loader = ConfigLoader("test/cfg")
    global_variables = config_loader.get_global_variables()
    global_variables["global_protocol_vars"] = [
        {"name": "email", "smtp_port": 587, "smtp_server": "smtp.gmail.com"}
    ]

    imported_task_def = config_loader.load_task_definition("email-transfer-dummy-plain")
    os.remove(task_definition_file)

    transfer_obj = transfer.Transfer(
        global_variables, "email-transfer-dummy-plain", imported_task_def
    )
    transfer_obj._set_remote_handlers()

    # Run the transfer
    assert transfer_obj.run()



# pylint: skip-file
# ruff: noqa

import json
import logging
import os
import subprocess

import pytest

from opentaskpy.cli import batch_validator

valid_batch_defininition = {
    "type": "batch",
    "tasks": [
        {"task_id": "task1", "order_id": 1},
        {"task_id": "task2", "order_id": 2, "dependencies": [1]},
    ],
}

valid_task_definition = {
    "type": "execution",
    "command": "echo 'hello world'",
    "protocol": "local",
}

batch_def_no_order_id_1 = {
    "type": "batch",
    "tasks": [
        {"task_id": "task1", "order_id": 2},
        {"task_id": "task2", "order_id": 3, "dependencies": [2]},
    ],
}

batch_def_dependency_non_existent_task_id = {
    "type": "batch",
    "tasks": [
        {"task_id": "task1", "order_id": 1},
        {"task_id": "task2", "order_id": 2, "dependencies": [3]},
    ],
}

batch_def_non_existent_task_id = {
    "type": "batch",
    "tasks": [
        {"task_id": "task3", "order_id": 1},
        {"task_id": "task4", "order_id": 2, "dependencies": [1]},
    ],
}


# Since the batch validator uses OTF_NOOP, we need to clear it down after
# the tests in this module have completed
@pytest.fixture(scope="module", autouse=True)
def tests_setup_and_teardown():
    # Will be executed before the first test
    old_environ = dict(os.environ)

    yield

    # Will be executed after the last test
    os.environ.clear()
    os.environ.update(old_environ)


@pytest.fixture(scope="function")
def write_config_files(tmpdir):
    # Write the valid_task_definition into a temporary dir as both task1 and task2
    for i in range(1, 3):
        with open(f"{tmpdir}/task{i}.json", "w", encoding="utf-8") as f:
            f.write(json.dumps(valid_task_definition))

    # Create a blank variables.json file in the temporary dir
    with open(f"{tmpdir}/variables.json", "w", encoding="utf-8") as f:
        f.write("{}")


def test_valid_batch_definition(tmpdir, write_config_files):

    # Write the valid_batch_definition into a temporary dir
    with open(f"{tmpdir}/test-task.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(valid_batch_defininition))

    # Run the main function and expect a True response
    assert batch_validator.main("test-task", 2, tmpdir)


def test_batch_definition_no_order_id_1(tmpdir, write_config_files):

    # Write the batch_def_no_order_id_1 into a temporary dir
    with open(f"{tmpdir}/test-task.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(batch_def_no_order_id_1))

    # Expect a False response as there's no task with order_id 1
    assert not batch_validator.main("test-task", 2, tmpdir)


def test_batch_definition_dependency_non_existent_task_id(tmpdir, write_config_files):

    # Write the batch_def_dependency_non_existent_task_id into a temporary dir
    with open(f"{tmpdir}/test-task.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(batch_def_dependency_non_existent_task_id))

    # Expect a False response as there's a dependency pointing at a non-existent task_id
    assert not batch_validator.main("test-task", 2, tmpdir)


def test_batch_definition_non_existent_task_id(tmpdir, write_config_files):

    # Write the batch_def_non_existent_task_id into a temporary dir
    with open(f"{tmpdir}/test-task.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(batch_def_non_existent_task_id))

    # Expect a FileNotFoundError exception as the task_id doesn't exist
    with pytest.raises(FileNotFoundError):
        batch_validator.main("test-task", 2, tmpdir)


def test_batch_validator_as_process(tmpdir, write_config_files):
    with open(f"{tmpdir}/test-task.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(valid_batch_defininition))

    # Call batch validator script as a sub process from the cli
    result = run_batch_validator("test-task", 2, tmpdir)
    assert result["returncode"] == 0

    # Test with invalid task and check return code is 1
    result = run_batch_validator("non-existent-task", 2, tmpdir)
    assert result["returncode"] == 1


def run_batch_validator(task, verbose="2", config_dir="test/cfg", noop=False):
    # We need to run the bin/task-run script to test this
    script = "python"
    args = [
        "src/opentaskpy/cli/batch_validator.py",
        "-t",
        task,
        "-v",
        str(verbose),
        "-c",
        config_dir,
    ]

    # Run the script
    result = subprocess.run([script] + args, capture_output=True)
    # Write stdout and stderr to the console
    logging.info("\n########## STDOUT ##########")
    logging.info(result.stdout.decode("utf-8"))
    # Get the console colour for red

    logging.info("########## STDERR ##########")
    logging.info(f"{result.stderr.decode('utf-8')}")

    return {
        "returncode": result.returncode,
        "stdout": result.stdout.decode("utf-8"),
        "stderr": result.stderr.decode("utf-8"),
    }



# pylint: skip-file
# ruff: noqa
import logging
import os
import re
import shutil
import time
from datetime import datetime

import opentaskpy.otflogging
from tests.fixtures.ssh_clients import *  # noqa: F403, F401


def test_log_redaction():
    redacted_string = opentaskpy.otflogging.REDACT_STRING

    # # Test a simple string
    log_message = "This is a test message"
    redacted_message = opentaskpy.otflogging.redact(log_message)
    assert redacted_message == "This is a test message"

    # Test a string with a sensitive key
    log_message = '{"password": "test1234", "username": "test"}'
    redacted_message = opentaskpy.otflogging.redact(log_message)
    assert (
        redacted_message
        == '{"password": "' + redacted_string + '", "username": "test"}'
    )

    log_message = '{"something": "-----BEGIN PRIVATE KEY-----\ntest1234\n-----END PRIVATE KEY-----"}'
    redacted_message = opentaskpy.otflogging.redact(log_message)
    assert redacted_message == '{"something": "' + redacted_string + '"}'

    log_message = '{"something": "-----BEGIN PRIVATE KEY-----\\ntest1234\\n-----END PRIVATE KEY-----"}'
    redacted_message = opentaskpy.otflogging.redact(log_message)
    assert redacted_message == '{"something": "' + redacted_string + '"}'

    # Without line breaks
    log_message = '{"something": "-----BEGIN PRIVATE KEY BLOCK-----test1234-----END PRIVATE KEY BLOCK-----"}'
    redacted_message = opentaskpy.otflogging.redact(log_message)
    assert redacted_message == '{"something": "' + redacted_string + '"}'


def test_define_log_file_name(env_vars, tmpdir, top_level_root_dir):
    # Pass in different task types and validate that the log file name is correct
    timestamp = datetime.now().strftime("%Y%m%d-") + r"\d{6}\.\d{3}"
    log_path = "logs"
    expected_result_regex = rf"{log_path}/no_task_id/{timestamp}_B_running.log"

    assert re.search(
        expected_result_regex, opentaskpy.otflogging._define_log_file_name(None, "B")
    )

    # Pass in a task ID and validate that the log file name is correct
    expected_result_regex = f"{log_path}/123/{timestamp}_B_running.log"
    assert re.search(
        expected_result_regex, opentaskpy.otflogging._define_log_file_name("123", "B")
    )

    # Manually override the log directory and validate that the path changes
    os.environ["OTF_LOG_DIRECTORY"] = str(tmpdir)
    expected_result_regex = rf"{tmpdir}/no_task_id/{timestamp}_B_running.log"
    assert re.search(
        expected_result_regex, opentaskpy.otflogging._define_log_file_name(None, "B")
    )

    # unset the environment variable
    del os.environ["OTF_LOG_DIRECTORY"]

    # Set a run prefix and check that log file names correlate
    os.environ["OTF_LOG_RUN_PREFIX"] = "test"
    expected_result = rf"{log_path}/no_task_id/test_B_running.log"
    assert opentaskpy.otflogging._define_log_file_name(None, "B") == expected_result

    # unset the environment variable if it's set
    if "OTF_LOG_RUN_PREFIX" in os.environ:
        del os.environ["OTF_LOG_RUN_PREFIX"]

    # Call the function to get the current name prefix, then call it again and validate the prefix is the same
    log_filename = opentaskpy.otflogging._define_log_file_name(None, "B")
    # Sleep 1 second to make sure the filename would have changed
    import time

    time.sleep(1)
    assert log_filename == opentaskpy.otflogging._define_log_file_name(None, "B")

    # Set an OTF_RUN_ID and validate that the directory name gets set to that
    os.environ["OTF_RUN_ID"] = "some-run-id"
    expected_result_regex = (
        rf"{log_path}/some-run-id/{timestamp}_B_some-task-id_running.log"
    )
    assert re.search(
        expected_result_regex,
        opentaskpy.otflogging._define_log_file_name("some-task-id", "B"),
    )

    # unset the environment variable if it's set
    if "OTF_RUN_ID" in os.environ:
        del os.environ["OTF_RUN_ID"]

    # Pass no task type, and expect None in it's place
    expected_result_regex = rf"{log_path}/no_task_id/{timestamp}_running.log"
    assert re.search(
        expected_result_regex, opentaskpy.otflogging._define_log_file_name(None, None)
    )


def test_init_logging(env_vars, top_level_root_dir):
    # Call init logging function and ensure that the returned logger includes a TaskFileHandler
    # pointing at the correct filename
    timestamp = datetime.now().strftime("%Y%m%d-") + r"\d{6}\.\d{3}"
    log_path = f"{top_level_root_dir}/logs"
    expected_result_regex = rf"{log_path}/some_task_id/{timestamp}_WRAPPER_running.log"
    logger = opentaskpy.otflogging.init_logging(
        "some.class.name1", task_id="some_task_id"
    )

    found_handler = False
    for log_handler in logger.handlers:
        if log_handler.__class__.__name__ == "TaskFileHandler":
            assert re.search(expected_result_regex, log_handler.baseFilename)
            found_handler = True
            break

    assert found_handler

    # Create a logger with a valid task type
    logger = opentaskpy.otflogging.init_logging(
        "some.class.name2", task_id="some_task_id", task_type="B"
    )
    expected_result_regex = rf"{log_path}/some_task_id/{timestamp}_B_running.log"
    # Find a handler of type TaskFileHandler in the logger
    found_handler = False
    for log_handler in logger.handlers:
        if log_handler.__class__.__name__ == "TaskFileHandler":
            assert re.search(expected_result_regex, log_handler.baseFilename)
            found_handler = True
            break

    assert found_handler

    # Disable logging via env variable and ensure there's no handler defined
    os.environ["OTF_NO_LOG"] = "1"
    logger = opentaskpy.otflogging.init_logging(
        "some.class.name3", task_id="some_task_id", task_type="B"
    )
    assert len(logger.handlers) == 0


def test_get_latest_log_file(env_vars):
    # Setup some dummy log files
    log_path = "test/testLogs"
    os.environ["OTF_LOG_DIRECTORY"] = log_path
    # Set debug logging
    opentaskpy.otflogging.logger.setLevel(logging.DEBUG)

    last_created_file = None

    # Empty the log directory
    shutil.rmtree(log_path, ignore_errors=True)

    # loop 1 to 10
    for _ in range(1, 11):
        log_file_name = opentaskpy.otflogging._define_log_file_name(None, "B")
        # Ensure the directory exists, if not, create it
        if not os.path.exists(os.path.dirname(log_file_name)):
            os.makedirs(os.path.dirname(log_file_name))

        # Create the file
        with open(log_file_name, "w") as f:
            f.write("test")
        # Sleep 1 second to make sure the filename would have changed
        time.sleep(0.01)
        # Clear the prefix which will have been set in the environment
        del os.environ["OTF_LOG_RUN_PREFIX"]
        last_created_file = log_file_name

    # Run the function to see what it thinks the latest file is,
    # this should return none, because they are all in the _running state to being with
    assert opentaskpy.otflogging.get_latest_log_file(None, "B") is None

    # Rename the last created file to remove the _running suffix
    os.rename(last_created_file, last_created_file.replace("_running", ""))
    last_created_file = last_created_file.replace("_running", "")

    # Check the rename worked
    assert os.path.exists(last_created_file)

    # Run the function again and validate that it still returns nothing
    assert opentaskpy.otflogging.get_latest_log_file(None, "B") is None

    # Rename this file to _failed
    failed_file = last_created_file.replace("_B", "_B_failed")
    os.rename(last_created_file, failed_file)
    # Check the rename worked
    assert os.path.exists(failed_file)

    # Run the function again and validate that it returns this file
    print("Testing finding a failed log to resume from")
    print(f"Expect to find {failed_file}")
    assert opentaskpy.otflogging.get_latest_log_file(None, "B") == failed_file

    # Create a new file that has succeeded, make sure it still returns None, as the latest
    # state is success
    log_file_name = opentaskpy.otflogging._define_log_file_name(None, "B")
    # Write to the file
    with open(log_file_name, "w") as f:
        f.write("test")
    # Rename the file to remove the _running suffix
    os.rename(log_file_name, log_file_name.replace("_running", ""))

    # Run the function again and validate that it returns nothing, as last state is success
    assert opentaskpy.otflogging.get_latest_log_file(None, "B") is None


def test_close_log_file(env_vars, tmpdir):
    os.environ["OTF_LOG_DIRECTORY"] = f"{tmpdir}/test/testLogs"

    # Create a logger and log something to it
    logger = opentaskpy.otflogging.init_logging(
        "some.class.name4", task_id="some_task_id", task_type="B"
    )
    logger.info("test")

    # Find a handler of type TaskFileHandler in the logger
    found_handler = None
    for log_handler in logger.handlers:
        logging.info(log_handler)
        if log_handler.__class__.__name__ == "TaskFileHandler":
            found_handler = log_handler
            break

    assert found_handler is not None

    # Now the file should exist, we will close it with a success state and then check it's been renamed
    opentaskpy.otflogging.close_log_file(logger, True)

    # Check the file has been renamed
    assert os.path.exists(found_handler.baseFilename.replace("_running", ""))

    # Now create a new logger and log something to it
    logger = opentaskpy.otflogging.init_logging(
        "some.class.name5", task_id="some_task_id", task_type="B"
    )
    logger.info("test")

    # Do the same, but make it fail
    opentaskpy.otflogging.close_log_file(logger, False)
    # Check the file has been renamed
    assert os.path.exists(
        logger.handlers[0].baseFilename.replace("_running", "_failed")
    )



"""File lookup plugin.

Reads the first line of a file and returns it as a string.

"""

import os

import opentaskpy.otflogging
from opentaskpy.exceptions import LookupPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

PLUGIN_NAME = "file"


def run(**kwargs) -> str:  # type: ignore[no-untyped-def]
    """Pull a variable from a named file.

    Args:
        **kwargs: Expect a kwarg named path. This should be the path to the file to read
        The file contents should be a single line, and will be returned as a string.

    Raises:
        LookupPluginError: Returned if the kwarg 'path' is not provided
        FileNotFoundError: Returned if the file does not exist

    Returns:
        _type_: The value read from the file
    """
    # Expect a kwarg named file
    if "path" not in kwargs:
        raise LookupPluginError(
            f"Missing kwarg: 'path' while trying to run lookup plugin '{PLUGIN_NAME}'"
        )

    # Check if the file exists
    if not os.path.isfile(kwargs["path"]):
        raise FileNotFoundError(
            f"File {kwargs['path']} does not exist while trying to run lookup plugin"
            f" '{PLUGIN_NAME}'"
        )

    # Read the file
    result = None
    with open(kwargs["path"], encoding="utf-8") as file_:
        result = file_.readline().strip()
        logger.log(12, f"Read '{result}' from file {kwargs['path']}")

    return result



# pylint: skip-file
import pytest

from opentaskpy.plugins.lookup.http_json import run


def test_http_json_plugin_missing_all():
    with pytest.raises(Exception) as ex:
        run()

    assert "Missing kwarg:" in (ex.value.args[0])


def test_http_json_plugin_missing_url():
    with pytest.raises(Exception) as ex:
        run(jsonpath="test")

    assert (
        ex.value.args[0]
        == "Missing kwarg: 'url' while trying to run lookup plugin 'http_json'"
    )


def test_http_json_plugin_missing_jsonpath():
    with pytest.raises(Exception) as ex:
        run(url="http://test.com")

    assert (
        ex.value.args[0]
        == "Missing kwarg: 'jsonpath' while trying to run lookup plugin 'http_json'"
    )


def test_http_json_plugin():
    # Run test with a valid URL and JSONPath
    result = run(
        url="https://jsonplaceholder.typicode.com/todos/1", jsonpath="$.userId"
    )
    assert result == 1


def test_http_json_plugin_complex():
    # A more complex JSONPath
    result = run(
        url="https://jsonplaceholder.typicode.com/todos", jsonpath="$[0].title"
    )
    assert result == "delectus aut autem"



"""Default filters provided by opentaskpy."""

import base64
import datetime


def now_localtime() -> datetime.datetime:
    """Alias of now."""
    return now()


def now() -> datetime.datetime:
    """Return the current time in the local timezone."""
    return datetime.datetime.now().astimezone()


def utc_now() -> datetime.datetime:
    """Alias of now_utc."""
    return now_utc()


def now_utc() -> datetime.datetime:
    """Return the current time in UTC."""
    return datetime.datetime.now(tz=datetime.UTC)


def delta_days(value: datetime.datetime, days: int) -> datetime.datetime:
    """Returns a new datetime object + or - the number of delta days.

    Args:
        value (datetime.datetime): Starting datetime object
        days (int): Days to increment or decrement the value

    Returns:
        datetime.datetime: New datetime object with the delta applied
    """
    return value + datetime.timedelta(days)


def delta_hours(value: datetime.datetime, hours: int) -> datetime.datetime:
    """Returns a new datetime object + or - the number of delta hours.

    Args:
        value (datetime.datetime): Starting datetime object
        hours (int): Hours to increment or decrement the value

    Returns:
        datetime.datetime: New datetime object with the delta applied
    """
    return value + datetime.timedelta(hours=hours)


def base64_encode(value: str) -> str:
    """Returns the base64 encoded value of a string.

    Args:
        value (str): The string to encode

    Returns:
        str: The base64 encoded string
    """
    return base64.b64encode(value.encode("utf-8")).decode("utf-8")


def base64_decode(value: str) -> str:
    """Returns the base64 decoded value of a string.

    Args:
        value (str): The string to decode

    Returns:
        str: The base64 decoded string
    """
    return base64.b64decode(value).decode("utf-8")



"""File caching plugin.

This is a basic caching plugin that when given a value, and path to a file, will update it with the new value
"""

import opentaskpy.otflogging
from opentaskpy.exceptions import CachingPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

CACHE_NAME = "file"


def run(**kwargs):  # type: ignore[no-untyped-def]
    """Update a file with a new value.

    Args:
        **kwargs: Expect kwargs named file, and value. This should be the file to write
        to, and the value to put into the file

    Raises:
        CachingPluginError: Returned if the kwarg 'file' or 'value' is not provided
        FileNotFoundException: Returned if unable to write to the destination
    """
    # Expect a kwarg named file
    expected_kwargs = ["file", "value"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise CachingPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run caching plugin"
                f" '{CACHE_NAME}'"
            )

    # Write the value to the file
    try:
        with open(kwargs["file"], "w", encoding="utf-8") as file_:
            file_.write(kwargs["value"])
            logger.log(12, f"Wrote '{kwargs['value']}' to file {kwargs['file']}")
    except FileNotFoundError as e:
        logger.error(f"Unable to write to file {kwargs['file']}. Error: {e}")
        raise e



# pylint: skip-file
# ruff: noqa
import json
import os
import random
from datetime import UTC, datetime, timedelta

import pytest
from freezegun import freeze_time
from jinja2.exceptions import UndefinedError
from pytest_shell import fs

from opentaskpy.config.loader import ConfigLoader
from opentaskpy.exceptions import VariableResolutionTooDeepError

GLOBAL_VARIABLES: str | None = None

RANDOM = random.randint(10000, 99999)


@pytest.fixture(scope="function")
def write_dummy_variables_file(tmpdir):
    # Write a nested variable to the global variables file
    json_obj = {
        "test": "{{ SOME_VARIABLE }}6",
        "SOME_VARIABLE": "{{ SOME_VARIABLE2 }}5",
        "SOME_VARIABLE2": "test1234",
        "NESTED_VAR_LEVEL_0": {
            "NESTED_VAR_LEVEL_1": {"NESTED_VAR_LEVEL_2": "nested_test1234"},
            "nested_variable_MIXED_CASE": "nested_mixed_case_test1234",
        },
    }

    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    # Unset any environment variables overrides
    if "OTF_VARIABLES_FILE" in os.environ:
        del os.environ["OTF_VARIABLES_FILE"]


def test_load_variables(tmpdir):
    # Ensure something satisfies the file lookup plugin
    fs.create_files(
        [
            {
                "/tmp/variable_lookup.txt": {
                    "content": "hello",
                }
            },
            {"/tmp/public_key_1.txt": {"content": "test1234"}},
            {"/tmp/public_key_2.txt": {"content": "test1234"}},
            {"/tmp/private_key_1.txt": {"content": "test1234"}},
            {"/tmp/private_key_2.txt": {"content": "test1234"}},
        ]
    )

    assert ConfigLoader("test/cfg") is not None


def test_load_variables_file_override(tmpdir):
    # Write a different variables file. Load it, and verify the variable matches
    json_obj = {
        "CUSTOM_VARS_FILE": "{{ XYZ }}1",
        "XYZ": "ABC",
    }

    fs.create_files(
        [
            {f"{tmpdir}/custom_vars_file.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    # Set environment variable to point to the new variables file
    os.environ["OTF_VARIABLES_FILE"] = f"{tmpdir}/custom_vars_file.json.j2"

    assert ConfigLoader("test/cfg") is not None

    # Verify the variable is loaded correctly
    assert ConfigLoader("test/cfg").get_global_variables()["CUSTOM_VARS_FILE"] == "ABC1"

    # Delete the file and verify that we get an exception thrown
    os.remove(f"{tmpdir}/custom_vars_file.json.j2")
    with pytest.raises(FileNotFoundError):
        ConfigLoader("test/cfg")


def test_load_multiple_variables_file_override(tmpdir):
    # Write a different variables file. Load it, and verify the variable matches
    custom_vars = {"CUSTOM_VARS_FILE": "{{ XYZ }}1", "XYZ": "ABC", "REPEATED": "123"}
    custom_vars2 = {"CUSTOM_VARS_FILE2": "{{ ZYX }}1", "ZYX": "DEF", "REPEATED": "456"}

    fs.create_files(
        [
            {
                f"{tmpdir}/custom_vars_file.json.j2": {
                    "content": json.dumps(custom_vars)
                }
            },
            {
                f"{tmpdir}/custom_vars_file2.json.j2": {
                    "content": json.dumps(custom_vars2)
                }
            },
        ]
    )

    # Set environment variable to point to the new variables files
    os.environ["OTF_VARIABLES_FILE"] = (
        f"{tmpdir}/custom_vars_file.json.j2,{tmpdir}/custom_vars_file2.json.j2"
    )

    assert ConfigLoader("test/cfg") is not None

    # Verify the variable is loaded correctly
    assert ConfigLoader("test/cfg").get_global_variables()["CUSTOM_VARS_FILE"] == "ABC1"
    assert (
        ConfigLoader("test/cfg").get_global_variables()["CUSTOM_VARS_FILE2"] == "DEF1"
    )

    # Verify that repeated variables has value loaded from last file loaded
    assert ConfigLoader("test/cfg").get_global_variables()["REPEATED"] == "456"

    # Delete the files and verify that we get an exception thrown
    os.remove(f"{tmpdir}/custom_vars_file.json.j2")
    os.remove(f"{tmpdir}/custom_vars_file2.json.j2")
    with pytest.raises(FileNotFoundError):
        ConfigLoader("test/cfg")


def test_load_task_definition(write_dummy_variables_file, tmpdir):
    # Initialise the task runner
    config_loader = ConfigLoader(tmpdir)
    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files([{f"{tmpdir}/task.json": {"content": '{"test": "{{ test }}"}'}}])

    expected_task_definition = {"test": "test123456"}

    # Test that the task definition is loaded correctly
    assert config_loader.load_task_definition("task") == expected_task_definition

    # Test that a non existent task definition file raises an error
    os.remove(f"{tmpdir}/task.json")
    with pytest.raises(FileNotFoundError) as e:
        config_loader.load_task_definition("task")
    assert e.value.args[0] == "Couldn't find task with name: task"


def test_lazy_loading_nested_variables(tmpdir):
    os.environ["OTF_LAZY_LOAD_VARIABLES"] = "1"
    # Create a variables file with some test variables in it
    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": '{"NEST": { "NEST1": "{{ now().strftime(\'%Y\') }}" } } '
                }
            },
        ]
    )

    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files(
        [{f"{tmpdir}/task.json": {"content": '{"test": "{{ NEST.NEST1 }}"}'}}]
    )

    evaluated_year = datetime.now().strftime("%Y")
    expected_task_definition = {"test": f"{evaluated_year}"}

    # Test that the task definition is loaded correctly
    config_loader = ConfigLoader(tmpdir)
    del os.environ["OTF_LAZY_LOAD_VARIABLES"]
    assert config_loader.load_task_definition("task") == expected_task_definition


def test_load_task_definition_lazy_load(tmpdir):
    # Set the OTF_LAZY_LOAD_VARIABLES environment variable
    os.environ["OTF_LAZY_LOAD_VARIABLES"] = "1"

    # Modify the variables file to include a variable that is templated
    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": '{"test": "{{ now().strftime(\'%Y\') }}"}'
                }
            }
        ]
    )

    # Evaluate the variable
    expected_variable = datetime.now().strftime("%Y")

    # Initialise the task runner
    config_loader = ConfigLoader(tmpdir)

    # Unset the env variable to not mess with other tests
    del os.environ["OTF_LAZY_LOAD_VARIABLES"]

    # Check the the test variable is not evaluated yet
    assert config_loader.global_variables["test"] == "{{ now().strftime('%Y') }}"

    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files([{f"{tmpdir}/task.json": {"content": '{"test": "{{ test }}"}'}}])

    expected_task_definition = {"test": f"{expected_variable}"}

    # Test that the task definition is loaded correctly
    assert config_loader.load_task_definition("task") == expected_task_definition

    # Test that a non existent task definition file raises an error
    os.remove(f"{tmpdir}/task.json")
    with pytest.raises(FileNotFoundError) as e:
        config_loader.load_task_definition("task")
    assert e.value.args[0] == "Couldn't find task with name: task"


def test_load_task_definition_lazy_load_too_deep(tmpdir):
    # Set the OTF_LAZY_LOAD_VARIABLES environment variable
    os.environ["OTF_LAZY_LOAD_VARIABLES"] = "1"

    # Create a task definition that uses a heavily nested variable
    fs.create_files(
        [{f"{tmpdir}/task.json": {"content": '{"test": "{{ SOME_VARIABLE7 }}"}'}}]
    )

    # Crete a variables file that uses a heavily nested variable
    json_obj = {
        "test": "{{ SOME_VARIABLE }}7",
        "SOME_VARIABLE": "{{ SOME_VARIABLE2 }}6",
        "SOME_VARIABLE2": "{{ SOME_VARIABLE3 }}5",
        "SOME_VARIABLE3": "{{ SOME_VARIABLE4 }}4",
        "SOME_VARIABLE4": "{{ SOME_VARIABLE5 }}3",
        "SOME_VARIABLE5": "{{ SOME_VARIABLE6 }}2",
        "SOME_VARIABLE6": "{{ SOME_VARIABLE7 }}1",
        "SOME_VARIABLE7": (
            "{{ SOME_VARIABLE8 }}{{ SOME_VARIABLE2 }}{{ SOME_VARIABLE3 }}"
        ),
        "SOME_VARIABLE8": "test1234",
    }

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    config_loader = ConfigLoader(tmpdir)

    del os.environ["OTF_LAZY_LOAD_VARIABLES"]

    with pytest.raises(VariableResolutionTooDeepError):
        config_loader.load_task_definition("task")


def test_load_new_variables_from_task_def(write_dummy_variables_file, tmpdir):
    config_loader = ConfigLoader(tmpdir)

    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files(
        [
            {
                f"{tmpdir}/task.json": {
                    "content": (
                        '{"test_var": "{{ test }}", "my_new_var": "{{ NEW_VARIABLE }}",'
                        ' "variables": {"NEW_VARIABLE": "NEW_VALUE"}}'
                    )
                }
            },
            {
                f"{tmpdir}/task1.json.j2": {
                    "content": (
                        '{"test_var": "{{ test }}","my_new_var": "{{ NEW_VARIABLE }}",'
                        ' "variables": {"NEW_VARIABLE": "NEW_VALUE"}}'
                    )
                },
            },
        ]
    )

    expected_task_definition = {
        "test_var": "test123456",
        "my_new_var": "NEW_VALUE",
        "variables": {"NEW_VARIABLE": "NEW_VALUE"},
    }

    # Test that the task definition is loaded correctly
    assert config_loader.load_task_definition("task") == expected_task_definition
    # task1 should fail because it's a jinja template and we don't support setting
    # variables in jinja templates
    config_loader = ConfigLoader(tmpdir)
    # Expect a jinja2 UndefinedError
    with pytest.raises(UndefinedError):
        config_loader.load_task_definition("task1")


def test_custom_plugin(tmpdir):
    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": '{"test": "{{ lookup(\'test_plugin\') }}"}'
                }
            },
        ]
    )
    # Symlink test/cfg/plugins to tmpdir/plugins
    os.symlink(
        os.path.join(os.path.dirname(__file__), "../test/cfg", "plugins"),
        f"{tmpdir}/plugins",
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == {"test": "hello"}


def test_default_filters(tmpdir):
    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": (
                        '{"test": "{{ \'SOME_TEXT\' | base64_encode }}", "test2": "{{ \'aGVsbG8=\' | base64_decode }}"}'
                    )
                }
            },
        ]
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == {
        "test": "U09NRV9URVhU",
        "test2": "hello",
    }


def test_custom_ordinal_date_filter(tmpdir):
    # Add the test config directory the the python path
    import sys

    sys.path.append(f"test/cfg/filters")
    import custom_ordinal_date as custom_ordinal_date

    test_data = [
        {
            "inputs": [[1, 4], datetime.strptime("2025-07-03", "%Y-%m-%d")],
            "expected": "25052",
        },
        {
            "inputs": [[1, 4], datetime.strptime("2025-07-04", "%Y-%m-%d")],
            "expected": "25053",
        },
        {
            "inputs": [[1, 4], datetime.strptime("2025-07-05", "%Y-%m-%d")],
            "expected": "25053",
        },
        {
            "inputs": [[1, 4], datetime.strptime("2025-07-08", "%Y-%m-%d")],
            "expected": "25054",
        },
    ]

    for test in test_data:
        assert (
            custom_ordinal_date.custom_weekday_ordinal(
                test["inputs"][0], test["inputs"][1]
            )
            == test["expected"]
        )


def test_custom_filter(tmpdir):
    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": '{"test": "{{ \'SOME_TEXT\' | md5 }}"}'
                }
            },
        ]
    )
    # Symlink test/cfg/filters to tmpdir/filters
    os.symlink(
        os.path.join(os.path.dirname(__file__), "../test/cfg", "filters"),
        f"{tmpdir}/filters",
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == {
        "test": "f7a5d4ca0b2a6cf667fe3da9fc06edff"
    }


def test_load_global_variables(tmpdir):
    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json": {"content": '{"test": "test1234"}'}},
        ]
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == {"test": "test1234"}
    os.remove(f"{tmpdir}/variables.json")

    # Load a .json.j2 file to check that works
    fs.create_files(
        [{f"{tmpdir}/variables.json.j2": {"content": '{"test": "test1234"}'}}]
    )

    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == {"test": "test1234"}
    os.remove(f"{tmpdir}/variables.json.j2")

    # Try loading a .json file that doesn't exist
    # Check this raise a FileNotFoundError
    with pytest.raises(FileNotFoundError):
        config_loader = ConfigLoader(tmpdir)

    # Create several files all in the same place, and ensure they all get
    # merged into one JSON object
    # Make several sub directories for various configs
    os.mkdir(f"{tmpdir}/variables1")
    os.mkdir(f"{tmpdir}/variables2")
    os.mkdir(f"{tmpdir}/variables3")

    fs.create_files(
        [
            {
                f"{tmpdir}/variables1/variables.json": {
                    "content": '{"test": "test1234"}'
                }
            },
            {
                f"{tmpdir}/variables2/variables.json": {
                    "content": '{"test2": "test5678"}'
                }
            },
            {
                f"{tmpdir}/variables3/variables.json": {
                    "content": '{"test3": "test9012"}'
                }
            },
        ]
    )
    config_loader = ConfigLoader(tmpdir)
    assert config_loader.get_global_variables() == {
        "test": "test1234",
        "test2": "test5678",
        "test3": "test9012",
    }


def test_resolve_templated_variables(tmpdir):
    json_obj = {"test": "{{ SOME_VARIABLE }}", "SOME_VARIABLE": "test1234"}
    json_resolved = {"test": "test1234", "SOME_VARIABLE": "test1234"}

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == json_resolved

    # Remove the variables file
    os.remove(f"{tmpdir}/variables.json.j2")

    # Test again, but with a nested variable
    json_obj = {
        "test": "{{ SOME_VARIABLE }}6",
        "SOME_VARIABLE": "{{ SOME_VARIABLE2 }}5",
        "SOME_VARIABLE2": "test1234",
    }

    json_resolved = {
        "test": "test123456",
        "SOME_VARIABLE": "test12345",
        "SOME_VARIABLE2": "test1234",
    }

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    assert config_loader.get_global_variables() == json_resolved

    # Remove the variables file
    os.remove(f"{tmpdir}/variables.json.j2")

    # Final test is to next 6 times, this should error as the limit is 5
    json_obj = {
        "test": "{{ SOME_VARIABLE }}7",
        "SOME_VARIABLE": "{{ SOME_VARIABLE2 }}6",
        "SOME_VARIABLE2": "{{ SOME_VARIABLE3 }}5",
        "SOME_VARIABLE3": "{{ SOME_VARIABLE4 }}4",
        "SOME_VARIABLE4": "{{ SOME_VARIABLE5 }}3",
        "SOME_VARIABLE5": "{{ SOME_VARIABLE6 }}2",
        "SOME_VARIABLE6": "{{ SOME_VARIABLE7 }}1",
        "SOME_VARIABLE7": (
            "{{ SOME_VARIABLE8 }}{{ SOME_VARIABLE2 }}{{ SOME_VARIABLE3 }}"
        ),
        "SOME_VARIABLE8": "test1234",
    }

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    # Test that the global variables are loaded correctly
    # Verify an exception with appropriate text is thrown
    with pytest.raises(Exception) as e:
        config_loader = ConfigLoader(tmpdir)

    assert e.value.args[0] == "Reached max depth of recursive template evaluation"


def test_resolve_lookups_in_task_definition(tmpdir):
    json_obj = {"test": "{{ SOME_VARIABLE }}", "SOME_VARIABLE": "test1234"}

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)

    # Now create a task definition file
    task_def = {
        "name": "test",
        "description": "{{ lookup('file', path='" + str(tmpdir) + "/lookup.txt') }}",
    }

    # Create a JSON file for the task definition
    fs.create_files(
        [
            {f"{tmpdir}/transfers/task_def.json": {"content": json.dumps(task_def)}},
            {f"{tmpdir}/lookup.txt": {"content": "FILE_LOOKUP_SUCCESS"}},
        ]
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)
    config_loader.get_global_variables()

    # Load the task definition
    task_def = config_loader.load_task_definition("task_def")
    # Check that the description has been resolved
    assert task_def["description"] == "FILE_LOOKUP_SUCCESS"


def test_default_date_variable_resolution(tmpdir):
    # Test that the default date variable is resolved correctly
    json_obj = {
        "YYYY": "{{ now().strftime('%Y') }}",
        "MM": "{{ now().strftime('%m') }}",
        "DD": "{{ now().strftime('%d') }}",
        "UTC_DD": "{{ utc_now().strftime('%d') }}",
        "MONTH_SHORT": "{{ now().strftime('%b') }}",
        "DAY_SHORT": "{{ now().strftime('%a') }}",
        "PREV_DD": "{{ (now()|delta_days(-1)).strftime('%d') }}",
        "PREV_UTC_DD": "{{ (utc_now()|delta_days(-1)).strftime('%d') }}",
        "PREV_NOW_UTC_DD": "{{ (now_utc()|delta_days(-1)).strftime('%d') }}",
        "PREV_MM": "{{ (now()|delta_days(-1)).strftime('%m') }}",
        "PREV_YYYY": "{{ (now()|delta_days(-1)).strftime('%Y') }}",
        "PREV_1H_HH": "{{ (now()|delta_hours(-1)).strftime('%H') }}",
        "PREV_1H_HH_LOCALTIME": (
            "{{ (now_localtime()|delta_hours(-1)).strftime('%H') }}"
        ),
    }

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    config_loader = ConfigLoader(tmpdir)
    config_loader.get_global_variables()

    # Check that the date variables are resolved correctly
    assert config_loader.get_global_variables()["YYYY"] == datetime.now().strftime("%Y")
    assert config_loader.get_global_variables()["MM"] == datetime.now().strftime("%m")
    assert config_loader.get_global_variables()["DD"] == datetime.now().strftime("%d")
    assert config_loader.get_global_variables()["UTC_DD"] == datetime.now(
        tz=UTC
    ).strftime("%d")
    assert config_loader.get_global_variables()[
        "MONTH_SHORT"
    ] == datetime.now().strftime("%b")
    assert config_loader.get_global_variables()["DAY_SHORT"] == datetime.now().strftime(
        "%a"
    )

    # Get datetime for yesterday
    yesterday = datetime.now() - timedelta(days=1)
    assert config_loader.get_global_variables()["PREV_YYYY"] == yesterday.strftime("%Y")
    assert config_loader.get_global_variables()["PREV_MM"] == yesterday.strftime("%m")
    assert config_loader.get_global_variables()["PREV_DD"] == yesterday.strftime("%d")
    assert config_loader.get_global_variables()[
        "PREV_NOW_UTC_DD"
    ] == yesterday.astimezone(tz=UTC).strftime("%d")

    # Get datetime for 1 hour ago
    previous_hour = datetime.now() - timedelta(hours=1)
    assert config_loader.get_global_variables()["PREV_1H_HH"] == previous_hour.strftime(
        "%H"
    )
    assert config_loader.get_global_variables()[
        "PREV_1H_HH_LOCALTIME"
    ] == previous_hour.strftime("%H")

    # Play around with the current time. Set it to a GMT time before the clocks change
    # Change the time to before BST starts
    initial_datetime = datetime(
        year=2024,
        month=3,
        day=31,
        hour=0,
        minute=59,
        second=59,
    )

    # Set an environment variable so that the current timezone is set to GMT0BST
    os.environ["TZ"] = "GMT0BST,M3.5.0/1,M10.5.0/2"

    with freeze_time(initial_datetime) as frozen_datetime:
        assert frozen_datetime() == initial_datetime

        config_loader = ConfigLoader(tmpdir)
        # At this point, "yesterday" should be 30th March 2024
        assert config_loader.get_global_variables()["PREV_DD"] == "30"

        frozen_datetime.tick()
        # Now we should be in BST, the previous day should still be 30th March 2024
        config_loader = ConfigLoader(tmpdir)
        assert config_loader.get_global_variables()["PREV_DD"] == "30"

        # Now jump to the next day at 00:59:59 again
        frozen_datetime.tick(delta=timedelta(seconds=-1, days=1))
        # This should now return 31st March 2024
        config_loader = ConfigLoader(tmpdir)
        assert config_loader.get_global_variables()["PREV_DD"] == "31"

        # Tick and check that the date is still correct
        frozen_datetime.tick()
        config_loader = ConfigLoader(tmpdir)
        assert config_loader.get_global_variables()["PREV_DD"] == "31"

    # Replicate the above test but with PREV_UTC_DD instead
    initial_datetime = datetime(
        year=2024,
        month=3,
        day=31,
        hour=0,
        minute=59,
        second=59,
    )

    with freeze_time(initial_datetime) as frozen_datetime:
        assert frozen_datetime() == initial_datetime

        config_loader = ConfigLoader(tmpdir)
        # At this point, "yesterday" should be 30th March 2024
        assert config_loader.get_global_variables()["PREV_UTC_DD"] == "30"

        frozen_datetime.tick()
        # Now we should be in BST, but we're looking at UTC the previous day should still be 30th March 2024
        config_loader = ConfigLoader(tmpdir)
        assert config_loader.get_global_variables()["PREV_UTC_DD"] == "30"

        # Now jump to the next day at 00:59:59 again
        frozen_datetime.tick(delta=timedelta(seconds=-1, days=1))
        # This should now return 31st March 2024
        config_loader = ConfigLoader(tmpdir)
        assert config_loader.get_global_variables()["PREV_UTC_DD"] == "31"

        # Tick and check that the date is still correct
        frozen_datetime.tick()
        config_loader = ConfigLoader(tmpdir)
        assert config_loader.get_global_variables()["PREV_UTC_DD"] == "31"


def test_override_date_variable_resolution(tmpdir):
    # Test that the default date variable is resolved correctly
    json_obj = {
        "YYYY": "{{ now().strftime('%Y') }}",
        "MM": "{{ now().strftime('%m') }}",
        "DD": "{{ now().strftime('%d') }}",
        "MONTH_SHORT": "{{ now().strftime('%b') }}",
        "DAY_SHORT": "{{ now().strftime('%a') }}",
        "PREV_DD": "{{ (now()|delta_days(-1)).strftime('%d') }}",
        "PREV_MM": "{{ (now()|delta_days(-1)).strftime('%m') }}",
        "PREV_YYYY": "{{ (now()|delta_days(-1)).strftime('%Y') }}",
    }

    # Create a JSON file with some test variables in it
    fs.create_files(
        [
            {f"{tmpdir}/variables.json.j2": {"content": json.dumps(json_obj)}},
        ]
    )

    os.environ["YYYY"] = "1901"
    os.environ["MM"] = "01"
    os.environ["DD"] = "14"

    config_loader = ConfigLoader(tmpdir)
    config_loader.get_global_variables()

    # Unset the variables
    del os.environ["YYYY"]
    del os.environ["MM"]
    del os.environ["DD"]

    # Check that the date variables are resolved correctly
    assert config_loader.get_global_variables()["YYYY"] == "1901"
    assert config_loader.get_global_variables()["MM"] == "01"
    assert config_loader.get_global_variables()["DD"] == "14"


def test_override_task_variables(tmpdir, write_dummy_variables_file):
    config_loader = ConfigLoader(tmpdir)

    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files(
        [
            {
                f"{tmpdir}/task.json": {
                    "content": (
                        '{"test_var": "{{ test }}", "variables": {"MY_VARIABLE":'
                        ' "value123"}}'
                    )
                }
            }
        ]
    )

    expected_task_definition = {
        "test_var": "test123456",
        "variables": {"MY_VARIABLE": "value123"},
    }

    # Test that the task definition is loaded correctly
    assert config_loader.load_task_definition("task") == expected_task_definition

    # Now override it with an environment variable and load it again
    os.environ["MY_VARIABLE"] = "overridden_value123"

    expected_task_definition = {
        "test_var": "test123456",
        "variables": {"MY_VARIABLE": "overridden_value123"},
    }
    new_task_definition = config_loader.load_task_definition("task")
    del os.environ["MY_VARIABLE"]
    assert new_task_definition == expected_task_definition


def test_override_nested_task_variables(tmpdir, write_dummy_variables_file):
    config_loader = ConfigLoader(tmpdir)

    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files(
        [
            {
                f"{tmpdir}/task.json": {
                    "content": (
                        '{"test_var": "{{ test }}", "variables": {"MY_VARIABLE": {"NESTED": "value123"}}}'
                    )
                }
            }
        ]
    )

    expected_task_definition = {
        "test_var": "test123456",
        "variables": {"MY_VARIABLE": {"NESTED": "value123"}},
    }

    # Test that the task definition is loaded correctly
    assert config_loader.load_task_definition("task") == expected_task_definition

    # Now override it with an environment variable and load it again
    os.environ["MY_VARIABLE.NESTED"] = "overridden_value123"

    expected_task_definition = {
        "test_var": "test123456",
        "variables": {"MY_VARIABLE": {"NESTED": "overridden_value123"}},
    }
    new_task_definition = config_loader.load_task_definition("task")
    del os.environ["MY_VARIABLE.NESTED"]
    assert new_task_definition == expected_task_definition


def test_override_nested_variables_file(tmpdir, write_dummy_variables_file):
    config_loader = ConfigLoader(tmpdir)

    # Create a task definition file (this isn't valid, but it proves if the evaluation of variables works)
    fs.create_files(
        [
            {
                f"{tmpdir}/task.json": {
                    "content": (
                        '{"test_var": "{{ NESTED_VAR_LEVEL_0.NESTED_VAR_LEVEL_1.NESTED_VAR_LEVEL_2 }}",'
                        '"test_var_mixed_case": "{{ NESTED_VAR_LEVEL_0.nested_variable_MIXED_CASE }}"}'
                    )
                }
            }
        ]
    )

    expected_task_definition = {
        "test_var": "nested_test1234",
        "test_var_mixed_case": "nested_mixed_case_test1234",
    }

    assert config_loader.load_task_definition("task") == expected_task_definition

    # Now override it with an environment variable and load it again
    os.environ["NESTED_VAR_LEVEL_0.NESTED_VAR_LEVEL_1.NESTED_VAR_LEVEL_2"] = (
        "overridden_value123"
    )
    os.environ["NESTED_VAR_LEVEL_0.nested_variable_MIXED_CASE"] = (
        "overridden_mixed_case_test1234"
    )

    # Evaluate the global variables
    global_vars = config_loader.get_global_variables()

    expected_task_definition = {
        "test_var": "overridden_value123",
        "test_var_mixed_case": "overridden_mixed_case_test1234",
    }

    new_task_definition = config_loader.load_task_definition("task")
    del os.environ["NESTED_VAR_LEVEL_0.NESTED_VAR_LEVEL_1.NESTED_VAR_LEVEL_2"]
    del os.environ["NESTED_VAR_LEVEL_0.nested_variable_MIXED_CASE"]
    assert new_task_definition == expected_task_definition


def test_override_task_specific_attribute(write_dummy_variables_file, tmpdir):
    # Define a basic scp transfer
    scp_task_definition = {
        "type": "transfer",
        "source": {
            "hostname": "172.16.0.11",
            "directory": "/tmp/testFiles/src",
            "fileRegex": ".*taskhandler.*\\.txt",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
        "destination": [
            {
                "hostname": "172.16.0.12",
                "directory": "/tmp/testFiles/dest",
                "protocol": {"name": "ssh", "credentials": {"username": "application"}},
            },
        ],
    }

    # Write the task definition to a file
    fs.create_files(
        [
            {
                f"{tmpdir}/transfers/test-task.json": {
                    "content": json.dumps(scp_task_definition)
                }
            }
        ]
    )

    # Override things
    os.environ["OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME"] = "non_existent_host"
    os.environ["OTF_OVERRIDE_TRANSFER_DESTINATION_0_HOSTNAME"] = "non_existent_host2"
    os.environ["OTF_OVERRIDE_TRANSFER_DESTINATION_0_PROTOCOL_CREDENTIALS_USERNAME"] = (
        "my_username"
    )

    # Load the task definition
    config_loader = ConfigLoader(tmpdir)
    task_definition = config_loader.load_task_definition("test-task")

    del os.environ["OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME"]
    del os.environ["OTF_OVERRIDE_TRANSFER_DESTINATION_0_HOSTNAME"]
    del os.environ["OTF_OVERRIDE_TRANSFER_DESTINATION_0_PROTOCOL_CREDENTIALS_USERNAME"]

    # Check that the hostname has been overridden
    assert task_definition["source"]["hostname"] == "non_existent_host"
    assert task_definition["destination"][0]["hostname"] == "non_existent_host2"
    assert (
        task_definition["destination"][0]["protocol"]["credentials"]["username"]
        == "my_username"
    )


def test_override_task_specific_attribute_execution(write_dummy_variables_file, tmpdir):

    execution_task_definition = {
        "type": "execution",
        "directory": "/tmp/testFiles/dest",
        "command": "echo 'hello world'",
        "protocol": {
            "name": "local",
            "some_attribute": "somethingrandom",
            "some_dict_attribute": {"some_dict_item": "somethingrandom"},
        },
    }

    # Write the task definition to a file
    fs.create_files(
        [
            {
                f"{tmpdir}/transfers/test-task-1.json": {
                    "content": json.dumps(execution_task_definition)
                }
            }
        ]
    )

    # Override things
    os.environ["OTF_OVERRIDE_EXECUTION_DIRECTORY"] = "/tmp/testFiles/dest2"
    os.environ["OTF_OVERRIDE_EXECUTION_PROTOCOL!!SOME_ATTRIBUTE"] = "my_attribute"
    os.environ[
        "OTF_OVERRIDE_EXECUTION_PROTOCOL!!SOME_DICT_ATTRIBUTE!!SOME_DICT_ITEM"
    ] = "my_dict_item"

    # Load the task definition
    config_loader = ConfigLoader(tmpdir)
    task_definition = config_loader.load_task_definition("test-task-1")

    del os.environ["OTF_OVERRIDE_EXECUTION_DIRECTORY"]
    del os.environ["OTF_OVERRIDE_EXECUTION_PROTOCOL!!SOME_ATTRIBUTE"]

    # Check that the directory and some_attribute properties have been overridden
    assert task_definition["directory"] == "/tmp/testFiles/dest2"
    assert task_definition["protocol"]["some_attribute"] == "my_attribute"
    assert (
        task_definition["protocol"]["some_dict_attribute"]["some_dict_item"]
        == "my_dict_item"
    )
